#include <stdio.h>

void classify(const float x[], float result[]);

int main() {
    float x[] = {{example}};
    float result[{{num_outputs}}];
    classify(x, result);
    return 0;
}

void classify(const float x[], float result[]) {
{{code}}
}
